;***************************************************************************************************
;*                                                                                                 *
;*             Flash Utility for updating the Monitor Code on the C02 Pocket SBC V1.01             *
;*                                                                                                 *
;*               Utility requires the new image is loaded into RAM located at $1000                *
;*                                                                                                 *
;*              This utility defaults to the new Monitor image size of 6KB only                    *
;*              The image is written to EEPROM starting at address $E000                           *
;*                                                                                                 *
;*      NOTE: A system coldstart is performed once the image has been written to EEPROM            *
;*            Utility uses BIOS routines to interact with User - I/O to console routines           *
;*                                                                                                 *
;***************************************************************************************************
;
;**************************************************************************************************
        PL      66      ;Page Length
        PW      132     ;Page Width (# of char/line)
        CHIP    W65C02S ;Enable WDC 65C02 instructions
        PASS1   OFF     ;Set ON for debugging
        INCLIST ON      ;Set ON for listing Include files
;**************************************************************************************************
;
        .INCLUDE        C02Constants4.asm   ;Include the Header for all equates, routines, etc.
;
;***************************************************************************************************
;**************************************************************************************************
;BIOS JUMP Table starts here:
;       - BIOS calls are listed below - total of 32
;       - Reserved calls are for future hardware support
;
B_IDE_RESET             .EQU    $FF00           ;Call 00
B_IDE_GET_STAT          .EQU    $FF03           ;Call 01
B_IDE_IDENTIFY          .EQU    $FF06           ;Call 02
B_IDE_READ_LBA          .EQU    $FF09           ;Call 03
B_IDE_WRITE_LBA         .EQU    $FF0C           ;Call 04
B_IDE_VERFY_LBA         .EQU    $FF0F           ;Call 05
B_IDE_SET_LBA           .EQU    $FF12           ;Call 06
B_IDE_SET_ADDR          .EQU    $FF15           ;Call 07
B_IDE_EN_CACHE          .EQU    $FF18           ;Call 08
B_IDE_DIS_CACHE         .EQU    $FF1B           ;Call 09
;
B_RTC_NVRD              .EQU    $FF1E           ;Call 10
B_RTC_NVWR              .EQU    $FF21           ;Call 11
B_RTC_INIT              .EQU    $FF24           ;Call 12
;
B_CHRIN2                .EQU    $FF27           ;Call 13
B_CHROUT2               .EQU    $FF2A           ;Call 14
;
B_CNT_INIT              .EQU    $FF2D           ;Call 15
B_CNT_STRT              .EQU    $FF30           ;Call 16
B_CNT_STOP              .EQU    $FF33           ;Call 17
;
B_CHRIN_NW              .EQU    $FF36           ;Call 18
B_CHRIN                 .EQU    $FF39           ;Call 19
B_CHROUT                .EQU    $FF3C           ;Call 20
;
B_SET_DLY               .EQU    $FF3F           ;Call 21
B_EXE_MSDLY             .EQU    $FF42           ;Call 22
B_EXE_LGDLY             .EQU    $FF45           ;Call 23
;
B_RESERVE               .EQU    $FF48           ;Call 24
;
B_INIT_VEC              .EQU    $FF4B           ;Call 25
B_INIT_CFG              .EQU    $FF4E           ;Call 26
B_INIT_28L92            .EQU    $FF51           ;Call 27
B_RESET_28L92           .EQU    $FF54           ;Call 28
;
B_WRMMNVEC0             .EQU    $FF57           ;Call 29
B_CLDMNVEC0             .EQU    $FF5A           ;Call 30
B_COLDSTRT              .EQU    $FF5D           ;Call 31
;
BIOS_MSG                .EQU    $FFD0           ;BIOS Startup Message is hard-coded here
;**************************************************************************************************
;
                .ORG    $0800
;
B_FLASH         ;BIOS Flash Utility start at $0800
;
                LDA     #$00            ;Get Flash Utility intro message
                JSR     PROMPT          ;Send to console
;
;Xfer byte write code to RAM for execution
                LDX     #BYTE_WRE-BYTE_WRS+1 ;Get length of byte write code
BYTE_XFER       LDA     BYTE_WRS-1,X    ;Get code
                STA     BURN_BYTE-1,X   ;Write code to RAM
                DEX                     ;Decrement index
                BNE     BYTE_XFER       ;Loop back until done
;
                LDA     #$01            ;Get Confirm message
                JSR     PROMPT          ;Send to console
                JSR     CONTINUE        ;Prompt User to continue (y/n)?
;
                LDA     #$02            ;Get Writing EEPROM message
                JSR     PROMPT          ;Send to console
;
OC_LOOP         LDA     OCNT_A          ;Check output buffer count
                BNE     OC_LOOP         ;Loop back until buffer sent
;
;Wait for 1/2 second for RAM/ROM access to settle
                LDA     #$32            ;Set milliseconds to 50(*10 ms)
                LDX     #$00            ;Zero out
                LDY     #$00            ;Zero out
                JSR     B_SET_DLY       ;Set Delay parameters
                JSR     B_EXE_MSDLY     ;Call delay for 1/2 second
;
                LDA     #$10            ;Get RAM start address MS location
                STA     SRCH            ;Store in Source high byte
                STZ     SRCL            ;Zero Source low byte
;
                LDA     #$F8            ;Get ROM start address high byte
                STA     TGTH            ;Store in Target high byte
                STZ     TGTL            ;Zero Target low byte
;
                LDA     #$06            ;Get upper value for 1.5KB ($0600)
                STA     LENH            ;Store in Length high byte
                STZ     LENL            ;Zero Length low byte
;
                JSR     MOVE_BLOCK      ;Move Block of data
;
                LDA     #$16            ;Get RAM start address MS location
                STA     SRCH            ;Store in Source high byte
                LDA     #$A0            ;Get RAM start address LS location
                STA     SRCL            ;Store in Source low byte
;
                LDA     #$FE            ;Get ROM start address high byte
                STA     TGTH            ;Store in Target high byte
                LDA     #$A0            ;Get ROM start address low byte
                STA     TGTL            ;Store in Target low byte
;
                LDA     #$01            ;Get upper value for 352 byte ($0160)
                STA     LENH            ;Store in Length high byte
                LDA     #$60            ;Get lower value for 352 bytes
                STA     LENL            ;Store in Length low byte
;
                JSR     MOVE_BLOCK      ;Move Block of data
;
                NOP
;
                JMP     B_COLDSTRT      ;Cold start SBC
;
MOVE_BLOCK      SEI                     ;Disable interrupts
;
COMPLP          LDA     LENL            ;Get low byte of length
                ORA     LENH            ;OR in High byte of length
                BEQ     QUITMV          ;If zero, nothing to compare/write
                JSR     BURN_BYTE       ;Else Burn a byte to EEPROM
                LDA     (SRCL)          ;Else load source
                CMP     (TGTL)          ;Compare to target
                BEQ     CMP_OK          ;If compare is good, continue
;
; Image write failed - just send a message and coldstart SBC
                LDA     #$03            ;Get Write failed message
                JSR     PROMPT          ;Send to console
                JMP     B_COLDSTRT      ;Cold start SBC
;
CMP_OK          JSR     UPD_STL         ;Update pointers
                BRA     COMPLP          ;Loop back until done
;
QUITMV          RTS                     ;Return to caller
;
;Routines to update pointers for memory operations. UPD_STL subroutine: Increments Source
; and Target pointers. UPD_TL subroutine: Increments Target pointers only, then drops into
; decrement Length pointer. Used by multiple Memory operation commands.
UPD_STL         INC     SRCL            ;Increment source low byte
                BNE     UPD_TL          ;Check for rollover
                INC     SRCH            ;Increment source high byte
UPD_TL          INC     TGTL            ;Increment target low byte
                BNE     DECLEN          ;Check for rollover
                INC     TGTH            ;Increment target high byte
;
;DECLEN subroutine: decrement 16-bit variable LENL/LENH
DECLEN          LDA     LENL            ;Get length low byte
                BNE     SKP_LENH        ;Test for LENL = zero
                DEC     LENH            ;Else decrement length high byte
SKP_LENH        DEC     LENL            ;Decrement length low byte
                RTS                     ;Return to caller
;
;Continue routine: called by commands to confirm execution, when No is confirmed, return address
;is removed from stack and the exit goes back to the Monitor input loop.
;Short version prompts for (Y/N) only.
CONTINUE        LDA     #$04            ;Get msg "cont? (Y/N)" to terminal
SH_CONT         JSR     PROMPT          ;Send to terminal
TRY_AGN         JSR     RDCHAR          ;Get keystroke from terminal
                CMP     #$59            ;"Y" key?
                BEQ     DOCONT          ;if yes, continue/exit
                CMP     #$4E            ;if "N", quit/exit
                BEQ     DONTCNT         ;Return if not ESC
                LDA     #$07            ;Get Beep code
                JSR     B_CHROUT        ;Send to console
                BRA     TRY_AGN         ;Loop back, try again
DONTCNT         PLA                     ;Else remove return address
                PLA                     ;discard it,
DOCONT          RTS                     ;Return
;
;PROMPT routine: Send indexed text string to terminal. Index is contained in A reg.
; String buffer address is stored in variable PROMPTL/PROMPTH. (placing here saves some space).
PROMPT          ASL     A               ;Multiply by two for msg table index
                TAY                     ;Xfer to index
                LDA     MSG_TABLE,Y     ;Get low byte address
                STA     PROMPTL         ;Store in Buffer pointer
                LDA     MSG_TABLE+1,Y   ;Get high byte address
                STA     PROMPTH         ;Store in Buffer pointer
;
PROMPT2         LDA     (PROMPTL)       ;Get string data
                BEQ     DOCONT          ;If null character, exit (borrowed RTS)
                JSR     B_CHROUT        ;Send character to terminal
                INC     PROMPTL         ;Increment low byte index
                BNE     PROMPT2         ;Loop back for next character
                INC     PROMPTH         ;Increment high byte index
                BRA     PROMPT2         ;Loop back and continue printing
;
BYTE_WRS
                LDA     #$AA            ;Get code $AA
                STA     EEPROM+$5555    ;Send to EEPROM
                LDA     #$55            ;Get code $55
                STA     EEPROM+$2AAA    ;Send to EEPROM
                LDA     #$A0            ;Get code $A0
                STA     EEPROM+$5555    ;Send to EEPROM
;
                LDA     (SRCL)          ;Get source byte
                STA     (TGTL)          ;Write to target byte
                LDA     (TGTL)          ;Read target byte (EEPROM)
                AND     #%01000000      ;Mask off bit 6 - toggle bit
BYTE_WLP        STA     TEMP3           ;Store in Temp location
                LDA     (TGTL)          ;Read target byte again (EEPROM)
                AND     #%01000000      ;Mask off bit 6 - toggle bit
                CMP     TEMP3           ;Compare to last read (toggles if write mode)
                BNE     BYTE_WLP        ;Branch back if not done
BYTE_WRE        RTS                     ;Return to caller
;
;RDCHAR subroutine: Waits for a keystroke to be entered.
; if keystroke is a lower-case alphabetical, convert it to upper-case
RDCHAR          JSR     B_CHRIN         ;Request keystroke input from terminal
                CMP     #$61            ;Check for lower case value range
                BCC     AOK             ;Branch if < $61, control code/upper-case/numeric
                SBC     #$20            ;Subtract $20 to convert to upper case
AOK             RTS                     ;Character received, return to caller
;
;Messages used
;
MSG_00  .DB     $0D,$0A
        .DB     "C02 BIOS Flash Utility Version 1.01"
        .DB     $0D,$0A
        .DB     "This will update the C02 BIOS code image"
        .DB     $0D,$0A
        .DB     "Make sure you have loaded the new BIOS image to $1000 before continuing"
        .DB     $0D,$0A
        .DB     "This utility will overwrite the 2KB BIOS space from $F800 to $FFFF"
        .DB     $0D,$0A
        .DB     $00
;
MSG_01  .DB     "Make sure you know what you are doing!"
        .DB     $0D,$0A
        .DB     "The C02 Pocket SBC will Coldstart when completed!"
        .DB     $0D,$0A
        .DB     $00
;
MSG_02  .DB     $0D,$0A
        .DB     "Writing BIOS image to EEPROM...."
        .DB     $0D,$0A
        .DB     "The C02 Pocket SBC will restart after the writing is completed!"
        .DB     $00
;
MSG_03  .DB     $0D,$0A,$0A
        .DB     "Error updating BIOS image!"
        .DB     $0D,$0A
        .DB     "Possible Write Protect Jumper or hardware error!"
        .DB     $0D,$0A
        .DB     "The C02 Pocket SBC will Coldstart now."
        .DB     $0D,$0A
        .DB     $00
;
MSG_04  .DB     "Would you like to continue? (y/n)"
        .DB     $00
;
MSG_TABLE       ;Message table: contains message addresses sent via the PROMPT routine
        .DW     MSG_00, MSG_01, MSG_02, MSG_03, MSG_04
;
        .END    ;End of User program

