;
;DOS/65 system interface module (SIM)
;Version 3.20 - RAM based
;
;This version is designed to work with the:
;C02 Pocket SBC with RTC/CF-Card adapter:
; - modified for 3.3V for MicroDrive, SC28L92 DUART and DS1511Y RTC
;
;Microdrive Disk configuration:
; a 8MB
; b 8MB
; c 8MB
; d 8MB
; e 8MB
; f 8MB
; g 8MB
; h 8MB
;
;This SIM version no longer uses Track/Sector parameters to access the Microdrive!
; - All calls and paraneters from PEM are now based on Record numbers.
; - Records are still set as 128-bytes per record.
; - Record number is a 16-bit value (0-65,535), so an 8MB limit still exists.
;
;The Disk Control Block parameter set is being modified slightly.
; This is done to accomodate the change from Track/Sector parameters to Records.
; The new DCB is configured as:
;
; Note: example shows an 8MB drive setup.
; - A total of 4096 blocks, where each block is 2048 bytes
; - the number of records per physical block is 4 (4*128=512)
; - the total number of reserved records is 256 (256*128=32,768 bytes)
; - Note: Reserved Records must be a multiple of 4!
;
;     .DW     4079            ;max block number
;     .DW     4               ;records per physical block (LBA)
;     .DW     256             ;number of reserved records (OS Boot area)
;     .DB     1               ;block size = 2048
;     .DW     1023            ;max directory
;     .DW     almpa           ;address of allocation map
;     .DB     128             ;do not do checksums
;     .DW     0               ;phony address
;
;A modified PEM module is required to use this new SIM version.
; PEM 3.20 also contains the changes for using Record numbers only!
;
;Note: This SIM version makes call to the C02BIOS directly!
; This results in a smaller code size and one less module to write and maintain.
;
;As this SIM supports a prototype C02 Pocket, the following devices exist:
; - Microdrive - supports 8 drives at 8MB each
; - Dual port UART - provides console and reader/punch functions
; - Realtime Clock - not currently used
;
;Note that this SIM version uses an additional 4-byte offset for LBA accesses!
; - This is required due to being a bootable system. The Microdrive setup uses a Partition Record
; - which will load the Boot Block from the active partition. The Boot Block and Boot Image are
; - NOT part of the allocated disk space that becomes the drives. This provides an additional
; - layer of safety and isolation. The Boot Block code will use the Load Header at the start of
; - CCM to provide the offset to allocated space and update the Partition Offset variable in
; - this SIM module (part_offset). This design allows the Boot Image to be easily moved to a new
; - location on the drive and not have to worry about the disk access. Only the Boot Block needs
; - to be updated.
;
;**************************************************************************************************
;Include required constants/equates/tables for SIM to assemble
;
        NOLIST
        INCLUDE         C02Constants4.asm       ;Constants/Equates - C02 BIOS/Monitor/Hardware
        INCLUDE         C02JMP_Table.asm        ;Jump Table entries for C02 BIOS/Monitor
        LIST
;
;**************************************************************************************************
;Fixed parameters
;
CLREOL          .EQU     1              ;clear to eol (ctrl-a)
CLREOS          .EQU     2              ;clear to eos (ctrl-b)
BS              .EQU     8              ;backspace (ctrl-h)
ESC             .EQU     $1B            ;escape character
;
;**************************************************************************************************
;
simram
;
;Main program, Start of SIM
;
sim             jmp     sim_boot        ;cold start SIM-DOS/65  simram+0
sim_wboote      jmp     sim_wboot       ;warm boot DOS/65       simram+3
                jmp     mon_consts      ;check for input        simram+6
                jmp     mon_conrde      ;get input              simram+9
                jmp     mon_conwrt      ;send to terminal       simram+12
                nop                     ;printer output         simram+15
                nop
                rts
                jmp     mon_punch       ;punch output           simram+18
                jmp     mon_reader      ;reader input           simram+21
                jmp     sim_home        ;set home drive record  simram+24
                jmp     sim_seldsk      ;select active disk     simram+27
                jmp     sim_setres      ;set reserved records   simram+30
                jmp     sim_setrec      ;set record number      simram+33
                jmp     sim_setdma      ;set buffer address     simram+36
                jmp     sim_read        ;read PEM record        simram+39
                jmp     sim_write       ;write PEM record       simram+42
                lda     #1              ;printer always ready   simram+45
                rts
                ldx     #128            ;clock entry (disabled) simram+48
                rts
                nop                     ;translate (obsolete)   simram+51
                nop
                rts
;
;console definition block
                byte    0               ;scratch
sysdef          byte    8               ;backspace
                byte    1               ;clear to end of line
                byte    $0C             ;forward space
                byte    0               ;normal video
                byte    '^'             ;invert video
                byte    24              ;lines per screen
                byte    80              ;char per line
                byte    $0C             ;form feed
                byte    $1E             ;home
                byte    2               ;clear to end of screen
;
;Opening id message
sim_opnmsg      .DB     CR,LF,"DOS/65 Beta Version!"
                .DB     CR,LF,"Version 3.20b",0
;
io_err_msg      .DB     CR,LF,"Disk I/O Error!",0
;
warm_msg        .DB     CR,LF,"*",0
;
sim_boot        jsr     b_ide_reset     ;Reset Microdrive
;                jsr     b_ide_dis_cache ;disable write caching
;
                lda     #<sim_opnmsg    ;Point to message
                ldy     #>sim_opnmsg    ;
                jsr     m_promptr       ;Send it via C02 Monitor
;
                ldx     #$FF            ;Set stack pointer
                txs                     ; to top of stack space
;
; To improve Disk I/O performance, the partition offset will be added
; to the drive offsets. This has to be a one-time calculation when
; SIM is started, otherwise the disk offset will be added each time
; SIM is reset, which doesn't work! This will take some additional
; memory space, but saves execution time in the get_lba routine.
; Once added in, the partition offset values below are zeroed out.
;
                ldx     #0              ;Zero index for drive offset
                ldy     #8              ;Zero index for 8 drives
drv_update_lp
                clc                     ;Clear carry for add
                lda     drv_offset+0,x  ;Get the drive offset
                adc     part_offset+0   ;Add the partition offset
                sta     drv_offset+0,x  ;Save it back
;
                lda     drv_offset+1,x  ;Get the drive offset
                adc     part_offset+1   ;Add the partition offset
                sta     drv_offset+1,x  ;Save it back
;
                lda     drv_offset+2,x  ;Get the drive offset
                adc     part_offset+2   ;Add the partition offset
                sta     drv_offset+2,x  ;Save it back
;
                inx                     ;Increment drive offset
                inx                     ;
                inx                     ;
                inx                     ;
;
                dey                     ;Decrement drive count
                bne     drv_update_lp   ;Loop back for 8 drives
;
; Now that all drive offsets have been updated with the partition offset,
; we will zero out the partition offset data (which is in RAM) so that it
; can't be added in again ;-) Ah... RAM based routines!
;
                stz     part_offset+0   ;Zero out Partition Offset
                stz     part_offset+1   ;
                stz     part_offset+2   ;
                stz     part_offset+3   ;
;
;Initialize iostat
                stz     iostat          ;Clear I/O status
;
;Set up jumps into dos/65 in page one
;
sim_setup
;First clear key disk access variables
                stz     sim_recflg      ;Clear record flag byte
;
sim_wsetup
                ldy     #6              ;Set index to 6
simset_lp       lda     sim_inttbl-1,y  ;Get SIM table data
                sta     $0100-1,y       ;Store into page $01 PEM area
                dey                     ;Decrement index
                bne     simset_lp       ;Loop until done
;
sim_finish
                lda     #<dflbuf        ;Get default buffer
                ldy     #>dflbuf        ; address
                jsr     sim_setdma      ;then set DMA for DOS/65
;
;Setup the BIOS for the default host buffer and block count.
; The BIOS reads or writes a single 512-byte block into it's own
; Host Buffer. The SIM code calls the C02 Pocket BIOS for all hardware
; access.
                lda     #<hstbuf        ;Get host buffer address
                ldy     #>hstbuf        ; hi byte
                ldx     #1              ;Block count =1
                jsr     b_ide_set_addr  ;Call BIOS
;
                jmp     ccmram          ;Then jump to CCM to finish startup
;
;Initialization table
sim_inttbl      .DB     $4C             ;JMP opcode
                .DW     sim_wboote      ;Warm Boot location
                .DB     $4C             ;JMP opcode
                .DW     pemram          ;PEM RAM location
;
;SIM Warm Boot - When any application exits, it should make a call to Warm Boot
; the system. That call will be handled here. As CCM can be over-written by any
; application to access additional memory if/when needed, the warm boot routine
; is responisble for reloading CCM. With DOS/65, both CCM and PEM are reloaded by
; default. As certain PEM routines are self-modifying and the PEM data is usually
; heavily altered during all operations, perhaps that's why Richard is specifying
; to reload it during a warm boot. SIM is NOT reloaded!
;
; IOSTAT and hence the default drive/user are not altered by Warm Boot.
;
sim_wboot
                lda     iostat          ;get current I/O status from PEM
                sta     sim_debug       ;save it for diags
;
; clear out the user space in Page Zero, this can cause crashes on warm start
; if leftovers from the last app are present.
; Apps like the Basic Compiler and Runtime are good examples.
;
                ldx     #$00            ;get index
clr_pg0
                stz     $00,x           ;clear page zero user space
                inx                     ;decrement index
                cpx     #$90            ;check for end
                bne     clr_pg0         ;loop back till done
;
                ldx     #$FF            ;set stack pointer
                txs                     ;to top of stack - safety
;
                lda     #<warm_msg      ;Point to message
                ldy     #>warm_msg      ;
                jsr     m_promptr       ;Send it
;
;As this is a bootable version of DOS/65, we need to reload CCM and PEM from disk.
; CCM, PEM and SIM are stored in the Reserved Records area of the bootable drive partition.
; Note that this also includes the drive allocations maps, which are zeroed. This eliminates
; the need to clear the allocation maps when the system is initially booted.
;
;The actual Cold Boot is handled ny the Boot Record, which is loaded from the Drive.
; This is the result of the Partition Loader loading and executing the active Boot Record.
; All of this is setup outside of DOS/65 and SIM. A separate SYS utility is used to setup
; the drive and load the DOS/65 code to the drive in the reserved area of the disk.
;
;To prevent the need to waste available drive space from any of the drives, there is 64KB of
; reserved space at the beginning of the active drive space. One 512-byte block is used for
; the Boot Record. This allows SIM to easily calculate the location for the operating system
; by using the first drive offset, subtract 64KB less 512 bytes (127 blocks) and then load
; CCM and PEM, which have a combined space of 5KB, or 10 512-byte blocks.
;
; for now, just get the parameters that are offset to the reserved area of the boot drive,
; which is a start of the drive offset. For now, we'll just use the first drive letter
; so no real calculations are needed... just point, load and go.
;
                lda     drv_offset+0    ;get LBA parameters
                ldy     drv_offset+1    ;
                ldx     drv_offset+2    ;
                jsr     b_ide_set_lba   ;call bios to set parameters
;
; next, set the load address for BIOS to load the CCM and PEM modules to
; the X reg contains the number of blocks to load. CCM starts at $B800, while
; PEM starts at $C100. SIM starts at $CC00, so the range is $B800 to $CBFF.
; This equates to 5KB, or 5*1024 divided by 512 = 10 blocks.
;
                lda     #<ramstrt       ;get load address lo
                ldy     #>ramstrt       ;get load address hi
                ldx     #10             ;get block count (5KB for CCM and PEM)
                jsr     b_ide_set_addr  ;call BIOS to set address and size
                jsr     b_ide_read_lba  ;call BIOS to load CCM/PEM to memory
                lda     ide_status_ram  ;get Status from BIOS call
                lsr     a               ;shift error bit to carry
                bcs     ide_wb_err      ;branch if error
;
                bra     sim_wsetup      ;warm setup for SIM
;
ide_wb_err
                lda     #<io_err_msg    ;point to error message
                ldy     #>io_err_msg
                jsr     m_promptr       ;send it
                jmp     sim_boot        ;restart SIM
;
;**************************************************************************************************
;
;Console Input and Output Routines
; There are three routines for Console I/O
; 1- CONSTS - Console Status: tests to see if there is a character available
; 2- CONRDE - Console Read: waits for a character to be available, then returns with it
; 3- CONWRT - Console Write: sends a character to the console (preserves data)
;
mon_consts
                lda     icnt_a          ;Load Input Buffer count
                beq     conste          ;Branch if Buffer empty
                dec     a               ;Else, Decrement A (LDA #$FF)
conste          rts                     ;Return to Caller
;
mon_conrde
                jsr     b_chrin         ;Call BIOS for character in (waits)
                and     #$7F            ;Strip off high bit (ASCII only)
                cmp     #bs             ;Check for Backspace character
                bne     conrdx          ;If not, branch
                lda     #delete         ;Else, replace with Delete code
conrdx          rts                     ;Return to Caller
;
mon_conwrt
                pha                     ;Save Character to Stack
                and     #$7F            ;Strip off high bit (ASCII only)
                cmp     #delete         ;Check for Delete code
                bne     notdlt          ;If not, branch
                lda     #' '            ;Else, load ASCII space
                bra     ntceos          ;branch and send
notdlt          cmp     #clreol         ;Check for EOL code
                bne     ntceol          ;If not, branch
                lda     #esc            ;Else, load ASCII escape
                jsr     b_chrout        ;Call BIOS for character send
                lda     #'T'            ;Load upper case "T"
                bra     ntceos          ;Branch to send/exit
ntceol          cmp     #clreos         ;Check for CLR code
                bne     ntceos          ;If not, branch
                lda     #esc            ;Load ASCII escape
                jsr     b_chrout        ;Call BIOS for character send
                lda     #'Y'            ;Load upper case "Y"
ntceos          jsr     b_chrout        ;Call BIOS for character send
endvid          pla                     ;Get original Character from Stack
                rts                     ;return to Caller
;
;**************************************************************************************************
;
;Additional routines for using the second serial port via C02BIOS routines
;
mon_punch
                jmp     b_chrout2       ;send data to serial port 2
;
mon_reader
                jmp     b_chrin2        ;get data from serial port 2
;
;**************************************************************************************************
;
;Start of Disk I/O Routines:
; Note that these routines rely on the C02BIOS calls for physical access of the Microdrive!
; 
; Blocking/De-blocking is always used for IDE devices (default LBA size = 512 bytes)
; All drives are configured with the same defaults as:
;
; - 4096 total blocks
; - 4 records per LBA (drive physical block size)
; - 16 records per block (DOS/65 Allocation size per extent)
;
; This section contains all new routines that are designed to work with Record Numbers!
;
; Records is a 16-bit field and record lengths are 128-bytes. The new routines provide a simple
; translation layer that will read and write blocks from the Microdrive. Each 512-byte block
; contains 4- 128-byte records. This calculates to a total of 8MB of disk space allocated via
; the 64K of records per drive letter.
;
; There are several routines which are required to support the SIM functions
;
;Record Move routines
; there are two sets of Record Move routines;
; - set one moves a Record from the LBA buffer to the DMA address specified by PEM
; - set two moves a Record from the DMA address specified by PEM to the LBA buffer
; In short, Set one is for reading a Record from disk and moving to memory, while
; Set two is for writing a Record from memory and moving to disk.
;
; In all cases, the LBA buffer is a 512-byte block and every Record is 128-bytes.
; this results in 4 Records per 512-byte block. The goal of these routines is to make
; them as quick as possible, even if it takes a bit of extra memory space.
;
;Home the selected disk
sim_home        bit     sim_recflg      ;test dirty block in record flag (4)
                bvs     no_home         ;if dirty block, exit (2/3)
                lda     #%00010000      ;get block write flag (2)
                trb     sim_recflg      ;reset write flag (6)
no_home         rts                     ;return (6)
;
;Select disk
sim_seldsk      and     #%00000111      ;mask for (0-7) eight drives only (2)
                sta     sim_seldrv      ;save as current drive (4)
                asl     a               ;multiply by two (2)
                tax                     ;make an index (2)
                lda     sim_dcbtbl,x    ;get DCB address (4)
                ldy     sim_dcbtbl+1,x  ; (4)
                rts                     ;and return (6)
;
;Table of dcb addresses.
; - All eight drives are in use for this new SIM version (3.20) - KM
;
sim_dcbtbl      .DW     dcb_a
                .DW     dcb_b
                .DW     dcb_c
                .DW     dcb_d
                .DW     dcb_e
                .DW     dcb_f
                .DW     dcb_g
                .DW     dcb_h
;
;Set DOS/65 memory buffer address
sim_setdma      sta     sim_dmaadr      ;save address low (4)
                sty     sim_dmaadr+1    ;save address high (4)
                rts                     ;return (6)
;
;Set Reserved Record Count
sim_setres      sta     sim_resrec      ;save reserved low (4)
                sty     sim_resrec+1    ;save reserved high (4)
                rts                     ;return (6)
;
;Set Requested Record
sim_setrec      sta     sim_pemrec      ;save record low (4)
                sty     sim_pemrec+1    ;save record high (4)
                rts                     ;return (6)
;
;*************************************************************************************************
;
;Read the selected DOS/65 record.
; note: (1,859 clock cycles to complete, does not include the cycle count from get_lba).
; at 8MHz, this is about 233 microseconds.
; note: minimum time for get_lba (correct LBA already loaded) is 152 clock cycles.
; This totals up to 2011 clock cycles, or about 252 microseconds to move a single
; 128-byte record. The works out to a max transfer rate of about 509KB/second.
;
; If a LBA needs to be loaded, I'll calculate the clock cycles later... much more work!
;
sim_read
                smb5    sim_recflg      ;set read op flag (5)
                jsr     get_lba         ;calculate/get LBA to read (6)
;
;Transfer the record from buffer to DMA
;
                lda     sim_recflg      ;get recflg (4)
                and     #%00000011      ;mask for lower 2-bits (2)
                asl     a               ;multiply by 2 for index (2)
                tax                     ;xfer to X reg as index (2)
;
                lda     buff_offset,x   ;get buffer offset address (4)
                ldy     buff_offset+1,x ; (4)
                sta     mover+1         ;save to routine lo (4)
                sty     mover+2         ;save to routine hi (4)
;
                lda     sim_dmaadr+0    ;get dma address lo (4)
                ldy     sim_dmaadr+1    ;get dma address hi (4)
                sta     mover+4         ;save to routine lo (4)
                sty     mover+5         ;save to routine hi (4)
                ldy     #0              ;zero index (2)
;
mover
                lda     $FFFF,y         ;get from source address (4)
                sta     $FFFF,y         ;move to destination address (5)
                iny                     ;increment offset (2)
                bpl     mover           ;loop back until done (2/3)
;
                rmb5    sim_recflg      ;reset read op flag (5)
                lda     #0              ;zero for no error (2)
                rts                     ;return (6)
;
;Write the selected DOS/65 record.
; Note: 1894 clock cycles to execute without an error.
; - does not include the clock cycles for get_lba or for writing a LBA via BIOS.
;
;
sim_write
                smb4    sim_recflg      ;set write op flag (5)
                jsr     get_lba         ;calculate/get LBA to write (6)
;
;Transfer the record DMA to buffer
;
                lda     sim_recflg      ;get recflg (4)
                and     #%00000011      ;mask for lower 2-bits (2)
                asl     a               ;multiply by 2 for index (2)
                tax                     ;xfer to X reg as index (2)
;
                lda     buff_offset,x   ;get buffer offset address (4)
                ldy     buff_offset+1,x ; (4)
                sta     movew+4         ;save to routine lo (4)
                sty     movew+5         ;save to routine hi (4)
;
                lda     sim_dmaadr+0    ;get dma address lo (4)
                ldy     sim_dmaadr+1    ;get dma address hi (4)
                sta     movew+1         ;save to routine lo (4)
                sty     movew+2         ;save to routine hi (4)
                ldy     #0              ;zero index (2)
;
movew
                lda     $FFFF,y         ;get from source address (4)
                sta     $FFFF,y         ;move to destination address (5)
                iny                     ;increment offset (2)
                bpl     movew           ;loop back until done (2/3)
;
                smb6    sim_recflg      ;set dirty block flag (5)
;
                jsr     b_ide_write_lba ;Call BIOS to Write LBA (6)
                lda     ide_status_ram  ;Get Status from BIOS call (3)
                lsr     a               ;Shift error bit to carry (2)
                bcs     ide_rw_err      ;Branch if error (2/3)
;
                lda     #%01010000      ;get dirty block & write flags (2)
                trb     sim_recflg      ;reset flags in sim_recflg (5)
                lda     #0              ;zero for no error (2)
                rts                     ;return to caller (6)
;
ide_rw_err
                lda     #<io_err_msg    ;Point to error message (2)
                ldy     #>io_err_msg    ; (2)
                jsr     m_promptr       ;Send it (6)
                jmp     sim_boot        ;Restart SIM (3)
;
buff_offset
                .dw     LBA_BUFFER+0    ;buffer offset 00
                .dw     LBA_BUFFER+128  ;buffer offset 01
                .dw     LBA_BUFFER+256  ;buffer offset 10
                .dw     LBA_BUFFER+384  ;buffer offset 11
;
;**************************************************************************************************
;
;Get LBA routine
; This is a core routine for SIM and all disk level activity. Now that PEM
; and SIM have been moved to a Record level access, the conversion between
; 128-byte records and 512-byte LBA block sizes can be streamlined a bit.
;
;This routine is common for both Reading and Writing of Records. As there
; is still a requirement to perform blocking/de-blocking between the two
; block sizes, this routine will perform the following tasks:
;
; 1- Calculate the Block Number from the Record number.
; 2- Save the Record offset within the calculated block number.
; 3- Check to see if a valid disk block is loaded in the buffer.
; 4- Perform a block read if there is no valid block loaded.
; 5- Test the valid block loaded against the required block number.
; 6- Test for a pending write for the valid block number.
; 7- Write a pending block and clear the dirty block flag if required.
; 8- Only load a block if needed, else return without disk I/O access.
; 9- Add the Drive offset to the required Block.
;
;This is a fair amount of work, but this core routine ensures we have the
; correct block in the default disk buffer on which to read or write
; records to/from. This routine also simplifies the Record Read and Write
; routines to focus only on handling the required 128-byte record.
; Another advantage is that this routine will only load a new block when
; it needs to, i.e., if the calculated block is already loaded, it will
; not reload it, but simply return and allow the Record Read and Write
; routines to carry out their record level function. This can improve
; the overall disk I/O performance.
;
;Calculate the block number (LBA) based on the Record number.
; The Record number is a 16-bit value (0-65,535) and is 128-bytes in size.
; To calculate the LBA, we simply divide it by 4, as there are 4 records
; per 512-byte block (LBA).
; We also save the lower 2-bits as this becomes the offset of the requested
; record into the block (as binary, 00,01,10,11).
;
;First, we need to take the Record Number and do the following:
; 1- Save the Record offset (the lower 2-bits of the Record number).
; 2- Perform a 16-bit Logical Right Shift by 2-bits (divide by 4).
; 3- Add the Drive number offset to the required LBA.
; 4- Save the calculated LBA, then check the flag byte for an LBA loaded.
;
;Second, we need to test for the following conditions:
; 1- If a LBA is loaded, check to see if it is the same LBA that we need.
; 2- If it is, just exit back to the calling routine, saving a disk read.
; 3- If it it NOT, check the dirty block flag in case it's been altered.
; 4- If the dirty block is set, write the current block and clear the flag,
;    then update the parameters and load the new LBA.
; 5- If the dirty block is not set, update the parameters and load the new LBA.
;
;The C02BIOS used here is Version 4.02.
; This BIOS supports 24-bit LBA addressing.
; Note: IDE uses 28-bit addressing, but the BIOS zeros the upper 4-bits,
; resulting in a limit of 8GB for the attached IDE device.
;
; The SIM configuration is set for 8 drives (A thru H), each at 8MB.
; any of the defined drives can have Reserved Records, but in practical
; use, only one drive should have Reserved Records to be use for holding
; the bootable image for the OS (CCM, PEM, SIM, Allocation Tables).
; The first drive will normally have the Reserved Records to hold the
; bootable OS image, where a total of 256 Records will equate to 32KB.
;
; NOTE: The PEM routine that calls SIM for the requested record also
; adds in the reserved record count, so we don't need to do that here!
;
;Once the Block number has been calculated from the Record number,
; an offset must be added which is based on the Drive number (0-7).
; Partition sizes are limited to 8MB
; Number of drives is limited to 8 (A - H)
; Maximum available Data is limited to 64MB
;
;The Offset table is 4 bytes per entry and adds an LBA count to
; shift the absolute LBA numbers to the appropriate drive number
; being acessed (16384 blocks per 8MB drive area).
;
;First, calculate the record number to LBA number and
; store the LBA number to sim_pemrec (3 bytes).
;
get_lba
                lda     #%00000011      ;Get mask for record offset bits (2)
                trb     sim_recflg      ;Reset record offset bits in flag (6)
;
                stz     sim_rectemp+2   ;Clear extended temp byte (4)
                lda     sim_pemrec+1    ;Get PEM Record number high byte )4)
                sta     sim_rectemp+1   ;Save to temp (4)
                lda     sim_pemrec+0    ;Get PEM Record number low byte (4)
                sta     sim_rectemp+0   ;Save to temp (4)
;
                and     #%00000011      ;Mask off upper 6 bits (2)
                ora     sim_recflg      ;OR into record flag (4)
                sta     sim_recflg      ;Update the record flag for new offset (4)
;
; now do the 16-bit shift to the right to divide by 4 and get the LBA number
;
                lsr     sim_rectemp+1   ;Shift high to right, LSB to carry (6)
                ror     sim_rectemp     ;Rotate low to right, MSB from carry (6)
                lsr     sim_rectemp+1   ;Shift high to the right, LSB to carry (6)
                ror     sim_rectemp     ;Rotate low to right, MSB from carry (6)
;
; the sim_pemrec pointer now has the logical LBA number requested by PEM.
; the sim_recflg also contains the record offset in the lower 2 bits,
; plus the other flags have been preserved (very important).
;
; Note that the partition offset (common for all drives) has been added as
; a one-time calculation during the first startup of SIM. As all drive
; offsets have been updated with the parition offset, we only need to
; apply the drive offset based on the current drive A-H (0-7).
;
                lda     sim_seldrv      ;Get active drive number (4)
                asl     a               ;Multiply by two (2)
                asl     a               ;Multiply by two (now = 4) (2)
                tax                     ;Xfer offset to X reg (2)
;
                clc                     ;Clear Carry for Add (2)
                lda     sim_rectemp+0   ;Get Low order LBA (4)
                adc     drv_offset+0,x  ;Add in Low order offset (4)
                sta     sim_rectemp+0   ;Save to Low order LBA (4)
;
                lda     sim_rectemp+1   ;Get Middle order LBA (4)
                adc     drv_offset+1,x  ;Add in Middle order offset (4)
                sta     sim_rectemp+1   ;Save to Middle order LBA (4)
;
                lda     sim_rectemp+2   ;Get High order LBA (4)
                adc     drv_offset+2,x  ;Add in High order offset (4)
                sta     sim_rectemp+2   ;Save to High order LBA (4)
;
; now, the actual required LBA has been calculated and is stored as a 24-bit value.
; this is stored in variables "sim_rectemp" (3 consecutive bytes).
; We now need to compare the actual LBA needed to the currently loaded LBA, if there
; is one. We first test for a valid LBA loaded by checking bit7 of the sim_recflg.
; note: if bit7 is set, the minus flag is also set.
;
                bit     sim_recflg      ;Test sim_recflg (4)
                bpl     need_lba        ;No valid LBA loaded, just go load it (2/3)
;
; we have determined that a valid LBA is loaded, so we need to compare the LBA
; that's been requested against the currently loaded LBA. If they are the same, we're
; done and can return. If they are not, then we need to check the dirty block flag to
; ensure the active block is written before we over-write it with the new one.
;
                 lda    sim_rectemp+0   ;Get low order requested LBA (4)
                 cmp    sim_curlba+0    ;Compare to current LBA (4)
                 bne    test_dirty      ;Bad compare, test for dirty block (2/3)
                 lda    sim_rectemp+1   ;Get middle order requested LBA (4)
                 cmp    sim_curlba+1    ;Compare to current LBA (4)
                 bne    test_dirty      ;Bad compare, test for dirty block (2/3)
                 lda    sim_rectemp+2   ;Get high order requested LBA (4)
                 cmp    sim_curlba+2    ;Compare to current LBA (4)
                 bne    test_dirty      ;Bad compare, test for dirty block (2/3)
                 rts                    ;Good compare, requested LBA is loaded (6)
;
; test the dirty block flag.
; the currently loaded LBA is not the one we need loaded.
; before we overwrite it, we'll test the dirty block flag.
; if it's set, we'll write that block to disk to update it, then transfer
; the parameters for the new LBA and load it.
;
test_dirty
                bit     sim_recflg      ;Bit test the sim_recflg (4)
                bvc     need_lba        ;Dirty flag not set, just load the new LBA (2/3)
;
; we have a dirty block in the buffer... so we need to write the current LBA back
; to the drive to ensure we don't get corrupted data on the drive!
; note that the existing BIOS parameters still point to the currently loaded LBA.
;
                jsr     b_ide_write_lba ;Call BIOS to Write LBA (6)
                lda     ide_status_ram  ;Get Status from BIOS call (3)
                lsr     a               ;Shift error bit to carry (2)
                bcs     get_lba_err     ;Branch if error (2/3)
                lda     #%01000000      ;Get bit6 - dirty block flag (2)
                trb     sim_recflg      ;Reset dirty block flag (6)
;
; now that the dirty block has been cleared, we can load the required LBA
; into the disk buffer, update the flags and return.
;
; if we branched here from above, no valid LBA is loaded, so we just transfer the
; LBA parameters to the BIOS and read the block into the default LBA buffer.
;
need_lba
                lda     sim_rectemp+0    ;Get required LBA parameters (Low byte) (4)
                ldy     sim_rectemp+1    ; Middle byte (4)
                ldx     sim_rectemp+2    ; High byte (4)
;
                sta     sim_curlba+0    ;Save as Current LBA
                sty     sim_curlba+1    ; for check the next
                stx     sim_curlba+2    ; time get_lba is called
;
                jsr     b_ide_set_lba   ;Call BIOS to set LBA parameters (6)
;
                jsr     b_ide_read_lba  ;Call BIOS to Read LBA (6)
                lda     ide_status_ram  ;Get Status from BIOS call (3)
                lsr     a               ;Shift error bit to carry (2)
                bcs     get_lba_err     ;Branch if error (2/3)
                lda     #%10000000      ;Get bit7 (2)
                tsb     sim_recflg      ;Set valid LBA flag bit (6)
                rts                     ;return to caller, we're done (6)
get_lba_err
                lda     #<io_err_msg    ;Point to error message (2)
                ldy     #>io_err_msg    ; (2)
                jsr     m_promptr       ;Send it (6)
                jmp     sim_boot        ;Restart SIM (3)
;
;**************************************************************************************************
;
;Table of LBA offsets for all drives A thru H (8MB each, 16KB Blocks per drive)
;
;  NOTE: These offsets are based on a non-partitioned drive!!! Relative to each other only!
;  The partition offset below as "part_offset" is added to the "drv_offset" values the first time
;  SIM is started. To ensure this only happens once, the "part_offset" data is zeroed out after it
;  is added to the "drv_offset". Once the new values are loaded, the active drive being accessed
;  is used as a base offset and the block offset calculated by the requested Record from PEM is
;  added to get the actual 512-byte block address on the drive.
;
drv_offset
                .LONG   0               ;Drive A default (w/OS boot image)
                .LONG   16384           ;Drive B
                .LONG   32768           ;Drive C
                .LONG   49152           ;Drive D
                .LONG   65536           ;Drive E
                .LONG   81920           ;Drive F
                .LONG   98304           ;Drive G
                .LONG   114688          ;Drive H
;
;Partition Offset
; this offset is patched in during boot time from the boot loader.
; this compensates for the drive partition being located anywhere on the drive.
;
part_offset
                .LONG   131072          ;Partition offset for drive letters
;
;**************************************************************************************************
;
;Disk Control Blocks (DCBs)
; Updated to support 8 drives by default - KM
;drive a (Microdrive 8MB)
dcb_a           .DW     4079            ;max block number-1
                .DW     4               ;records per physical block
                .DW     256             ;number of reserved records (32KB OS-boot)
                .DB     1               ;block size = 2048
                .DW     1023            ;max directory-1
                .DW     almpa           ;address of allocation map
                .DB     128             ;do not do checksums
                .DW     0               ;phony address
;drive b (Microdrive 8MB)
dcb_b           .DW     4095            ;max block number-1
                .DW     4               ;records per physical block
                .DW     0               ;number of reserved records
                .DB     1               ;block size = 2048
                .DW     1023            ;max directory-1
                .DW     almpb           ;address of allocation map
                .DB     128             ;do not do checksums
                .DW     0               ;phony address
;drive c (Microdrive 8MB)
dcb_c           .DW     4095            ;max block number-1
                .DW     4               ;records per physical block
                .DW     0               ;number of reserved records
                .DB     1               ;block size = 2048
                .DW     1023            ;max directory-1
                .DW     almpc           ;address of allocation map
                .DB     128             ;do not do checksums
                .DW     0               ;phony address
;drive d (Microdrive 8MB)
dcb_d           .DW     4095            ;max block number-1
                .DW     4               ;records per physical block
                .DW     0               ;number of reserved records
                .DB     1               ;block size = 2048
                .DW     1023            ;max directory-1
                .DW     almpd           ;address of allocation map
                .DB     128             ;do not do checksums
                .DW     0               ;phony address
;drive e (Microdrive 8MB)
dcb_e           .DW     4095            ;max block number-1
                .DW     4               ;records per physical block
                .DW     0               ;number of reserved records
                .DB     1               ;block size = 2048
                .DW     1023            ;max directory-1
                .DW     almpe           ;address of allocation map
                .DB     128             ;do not do checksums
                .DW     0               ;phony address
;drive f (Microdrive 8MB)
dcb_f           .DW     4095            ;max block number-1
                .DW     4               ;records per physical block
                .DW     0               ;number of reserved records
                .DB     1               ;block size = 2048
                .DW     1023            ;max directory-1
                .DW     almpf           ;address of allocation map
                .DB     128             ;do not do checksums
                .DW     0               ;phony address
;drive g (Microdrive 8MB)
dcb_g           .DW     4095            ;max block number-1
                .DW     4               ;records per physical block
                .DW     0               ;number of reserved records
                .DB     1               ;block size = 2048
                .DW     1023            ;max directory-1
                .DW     almpg           ;address of allocation map
                .DB     128             ;do not do checksums
                .DW     0               ;phony address
;drive h (Microdrive 8MB)
dcb_h           .DW     4095            ;max block number-1
                .DW     4               ;records per physical block
                .DW     0               ;number of reserved records
                .DB     1               ;block size = 2048
                .DW     1023            ;max directory-1
                .DW     almph           ;address of allocation map
                .DB     128             ;do not do checksums
                .DW     0               ;phony address
;
;**************************************************************************************************
;
; SIM Data area
;
;Record level data area required by SIM
sim_seldrv      defs    1               ;Current selected drive number (0-7)
sim_dmaadr      defs    2               ;DMA address to Record transfer
sim_resrec      defs    2               ;Reserved Record count
sim_pemrec      defs    2               ;Record number requested by PEM
sim_rectemp     defs    3               ;Calculated LBA required by PEM (with Partition/Drive offsets)
sim_curlba      defs    3               ;LBA number currently loaded in Buffer (plus warm boot LBA)
sim_debug       defs    1               ;extra data variable for debugging ;-)
;
;**************************************************************************************************
;
                end
